/**************************************************************************

	ns

	script_exec.c

	(C) Copyright 1996-2001 By Tomoaki Nakashima. All right reserved.
		http://www.nakka.com/
		nakka@nakka.com

**************************************************************************/

/**************************************************************************
	Include Files
**************************************************************************/

#include <windows.h>
#include <tchar.h>

#ifndef _WIN32_WCE
#include <process.h>
#endif

#include "script.h"
#include "script_regex.h"
#include "script_string.h"
#include "script_memory.h"


/**************************************************************************
	Define
**************************************************************************/

#define RET_EXIT				-3
#define RET_RETURN				-2
#define RET_ERROR				-1
#define RET_SUCCESS				0
#define RET_BREAK				1
#define RET_CONTINUE			2

#define FUNC_SCRIPT				1
#define FUNC_LIBRARY			2

#define ARGUMENT_ADDRESS		'&'
#define ARGUMENT_VARIABLE		"arg"
#define ARGUMENT_EXEC			"?"
#define ALL_ARGUMENT_EXEC		"all."
#define THREAD_EXEC				'&'
#define STD_FUNC_HEAD			"_sfunc_"
#define SCRIPT_SEP				'.'
#define VARIABLE_CHAR			'%'


/**************************************************************************
	Global Variables
**************************************************************************/

static CRITICAL_SECTION cs;
static int FuncCnt = 0;
static BOOL TerminateFlag;


/**************************************************************************
	Local Function Prototypes
**************************************************************************/

static SCRIPTINFO *GetScriptInfo(SCRIPTINFO *sci, char *func);

static VALUEINFO *Index2Hash(EXECINFO *ei, VALUEINFO *pvi, int index);
static VALUEINFO *DeclVariable(EXECINFO *ei, char *name, char *err);
static VALUEINFO *FindValueInfo(VALUEINFO *vi, char *name);
static VALUEINFO *AddValueInfo(VALUEINFO **vi_root, char *name, VALUE *v);

static int GetStringConvLen(EXECINFO *ei, char *fmt, VALUEINFO *vi);
static char *StringConv(EXECINFO *ei, char *fmt);

static TOKEN *FindCase(EXECINFO *ei, TOKEN *cu_tk, VALUEINFO *v1);
static VALUEINFO *UnaryCalcValue(EXECINFO *ei, VALUEINFO *vi, int c);
static VALUEINFO *IntegerCalcValue(EXECINFO *ei, VALUEINFO *v1, VALUEINFO *v2, int c);
static VALUEINFO *StringCalcValue(EXECINFO *ei, VALUEINFO *v1, VALUEINFO *v2, int c);
static VALUEINFO *HashCalcValue(EXECINFO *ei, VALUEINFO *v1, VALUEINFO *v2, int c);
static VALUEINFO *CalcValue(EXECINFO *ei, VALUEINFO *v1, VALUEINFO *v2, int c);

static TOKEN *ExpandArgument(EXECINFO *ei, TOKEN *tk, VALUEINFO *param);
static void CopyAllArgument(EXECINFO *ei, VALUEINFO *param);

static VALUEINFO *ExecNameFunction(SCRIPTINFO *csci, FUNCINFO *fi, VALUEINFO *param, long lParam);
static int ExecArgumentFunction(SCRIPTINFO *sci, VALUEINFO *param, int *exit, long lParam);
static int ExecArgumentAllFunction(SCRIPTINFO *sci, VALUEINFO *param, long lParam);
static VALUEINFO *ExecLibFunction(EXECINFO *ei, LIBFUNC StdFunc, char *name, VALUEINFO *param);
static void SetExecFuncCnt(int i);
static unsigned int CALLBACK ScriptThread(SCRIPT_THREAD_INFO *sci);
static void SetGlobalFlag(VALUEINFO *vi);


/******************************************************************************

	InitializeScript

	XNvg̏

******************************************************************************/

void InitializeScript()
{
	InitializeCriticalSection(&cs);

#ifdef NO_EXPORT
	//֐e[ȕ
	InitFuncAddress();
#endif

}


/******************************************************************************

	EndScript

	XNvg̏I

******************************************************************************/

void EndScript()
{
	DeleteCriticalSection(&cs);
}


/******************************************************************************

	TerminateScript

	XNvg̋I

******************************************************************************/

void TerminateScript()
{
	ScriptWait();
	TerminateFlag = TRUE;
	ScriptUnWait();
	while(1){
		Sleep(0);

		ScriptWait();
		if(FuncCnt <= 0){
			ScriptUnWait();
			break;
		}
		ScriptUnWait();
	}
	ScriptWait();
	TerminateFlag = FALSE;
	ScriptUnWait();
}


/******************************************************************************

	ScriptWait

	ݒ

******************************************************************************/

void ScriptWait()
{
	EnterCriticalSection(&cs);
}


/******************************************************************************

	ScriptUnWait

	

******************************************************************************/

void ScriptUnWait()
{
	LeaveCriticalSection(&cs);
}


/******************************************************************************

	CheckEndEvent

	XNvgICxg̃`FbN

******************************************************************************/

BOOL CheckEndEvent()
{
	BOOL ret;

	ScriptWait();
	ret = TerminateFlag;
	ScriptUnWait();
	return ret;
}


/******************************************************************************

	Error

	G[

******************************************************************************/

void Error(EXECINFO *ei, char *buf, char *msg, BOOL trace)
{
#define ERR_HEAD	"### Error: "
#define ERR_TRACE	"        at "
	EXECINFO *pei = ei;
	VALUEINFO *vi, *pvi;
	LIBFUNC StdFunc;
	char *p, *r, *s, *t, *f;
	char ErrStr[BUF_SIZE];
	int i;

	if(ei == NULL || CheckEndEvent() == TRUE){
		return;
	}

	for(; pei->parent != NULL; pei = pei->parent);

	//bZ[W
	if(trace == FALSE){
		f = p = alloc_join(ERR_HEAD, buf);
		p = alloc_join(p, " : ");
		mem_free(&f);f = p;
	}else{
		f = p = alloc_copy(ERR_TRACE);
	}
	//XNvg
	p = alloc_join(p, pei->sci->name);
	mem_free(&f);f = p;
	if(pei->sci != NULL && pei->sci->name != NULL && pei->Name != NULL){
		p = alloc_join(p, ".");
		mem_free(&f);f = p;
	}
	//֐
	p = alloc_join(p, pei->Name);
	mem_free(&f);f = p;
	if(msg != NULL && pei->sci->buf != NULL && msg >= pei->sci->buf && msg <= (pei->sci->buf + tstrlen(pei->sci->buf))){
		//s
		i = 1;
		for(s = r = pei->sci->buf; r < msg; r++){
			if(*r == '\n'){
				s = r;
				i++;
			}
		}
		//sԍ
		p = alloc_join(p, ":(");
		mem_free(&f);f = p;
		t = i2a(i);
		p = alloc_join(p, t);
		mem_free(&f);f = p;
		mem_free(&t);
		p = alloc_join(p, ") : ");
		mem_free(&f);f = p;

		//
		for(; *s == '\r' || *s == '\n' || *s == '\t' || *s == ' '; s++);
		for(r = s; *r != '\0' && *r != '\r' && *r != '\n'; r++);
		t = alloc_copy_n(s, r - s + 1);
		p = alloc_join(p, t);
		mem_free(&f);f = p;
		mem_free(&t);

	}else if(msg != NULL){
		//bZ[W
		p = alloc_join(p, " : ");
		mem_free(&f);f = p;
		p = alloc_join(p, msg);
		mem_free(&f);f = p;
	}

	//G[bZ[W̏o
#ifdef NO_EXPORT
	StdFunc = GetFuncAddress("_sfunc_error");
#else
	StdFunc = GetProcAddress(GetModuleHandle(NULL), "_sfunc_error");
#endif
	if(StdFunc == NULL){
		return;
	}
	vi = AllocValue();
	if(vi == NULL){
		return;
	}
	pvi = AllocValue();
	if(pvi == NULL){
		return;
	}
	pvi->v->u.Value = p;
	pvi->v->Type = TYPE_STRING;
	*ErrStr = '\0';
	StdFunc(pei, pvi, &vi, ErrStr);
	FreeValue(vi);
	FreeValue(pvi);
}


/******************************************************************************

	FreeToken

	g[N̉

******************************************************************************/

void FreeToken(TOKEN *tk)
{
	if(tk == NULL) return;
	FreeToken(tk->target);
	FreeToken(tk->next);

	if(tk->nfa != NULL){
		free_nfa(tk->nfa);
	}

	mem_free(&(tk->buf));
	mem_free(&tk);
}


/******************************************************************************

	FreeValue

	ϐ̉

******************************************************************************/

void FreeValue(VALUEINFO *vi)
{
	if(vi == NULL || (int)vi == -1) return;
	mem_free(&(vi->Name));
	if(vi->v != NULL && vi->v->vi == vi){
		if(vi->v->Type == TYPE_STRING){
			mem_free(&(vi->v->u.Value));
		}
		mem_free(&(vi->v));
		FreeValueList(vi->vi);
	}
	mem_free(&vi);
}


/******************************************************************************

	FreeValueList

	ϐXg̉

******************************************************************************/

void FreeValueList(VALUEINFO *vi)
{
	VALUEINFO *tmpvi;

	if((int)vi == -1) return;
	while(vi != NULL){
		tmpvi = vi->next;
		FreeValue(vi);
		vi = tmpvi;
	}
}


/******************************************************************************

	FreeExecInfo

	s̉

******************************************************************************/

void FreeExecInfo(EXECINFO *ei)
{
	if(ei == NULL) return;
	FreeValueList(ei->inc_vi);
	FreeValueList(ei->dec_vi);
	FreeValueList(ei->vi);
	ei->vi = NULL;
	ei->Name = NULL;
}


/******************************************************************************

	FreeFuncInfo

	֐̉

******************************************************************************/

void FreeFuncInfo(FUNCINFO *fi)
{
	if(fi == NULL) return;
	FreeFuncInfo(fi->next);

	mem_free(&(fi->Name));
	mem_free(&fi);
}


/******************************************************************************

	FreeLibInfo

	Củ

******************************************************************************/

void FreeLibInfo(LIBRARYINFO *lib)
{
	if(lib == NULL) return;
	FreeLibInfo(lib->next);

	FreeLibrary(lib->hModul);
	mem_free(&lib);
}


/******************************************************************************

	FreeScriptInfo

	XNvg̉

******************************************************************************/

void FreeScriptInfo(SCRIPTINFO *sci)
{
	if(sci == NULL) return;
	FreeScriptInfo(sci->next);

	mem_free(&sci->name);
	mem_free(&sci->path);
	mem_free(&sci->buf);

	FreeToken(sci->tk);
	FreeValueList(sci->vi);
	FreeFuncInfo(sci->fi);
	FreeLibInfo(sci->lib);

	if(sci == sci->sci_top){
		DeleteCriticalSection(&sci->cs);
	}
	mem_free(&sci);
}


/******************************************************************************

	InitializeScriptInfo

	XNvg̏

******************************************************************************/

void InitializeScriptInfo(SCRIPTINFO *sci, BOOL op_exp)
{
	ZeroMemory(sci, sizeof(SCRIPTINFO));
	sci->sci_top = sci;
	sci->explicit_val = op_exp;
	//IuWFNg
	InitializeCriticalSection(&sci->cs);
}


/******************************************************************************

	GetScriptInfo

	XNvg

******************************************************************************/

static SCRIPTINFO *GetScriptInfo(SCRIPTINFO *sci, char *func)
{
	SCRIPTINFO *csci = sci->sci_top;
	char *name;
	char *r;

	if(csci == NULL){
		return NULL;
	}

	for(r = func; *r != '\0' && *r != SCRIPT_SEP; r++);
	if(*r != SCRIPT_SEP){
		return NULL;
	}
	name = alloc_copy_n(func, r - func + 1);

	for(csci = csci->next; csci != NULL; csci = csci->next){
		if(csci->name != NULL && str_cmp_i(name, csci->name) == 0){
			mem_free(&name);
			return csci;
		}
	}
	mem_free(&name);
	return NULL;
}


/******************************************************************************

	GlobalWait

	ݒ

******************************************************************************/

void GlobalWait(SCRIPTINFO *sci)
{
	EnterCriticalSection(&sci->cs);
}


/******************************************************************************

	GlobalUnWait

	

******************************************************************************/

void GlobalUnWait(SCRIPTINFO *sci)
{
	LeaveCriticalSection(&sci->cs);
}


/******************************************************************************

	AllocValue

	ϐ̊m

******************************************************************************/

VALUEINFO *AllocValue()
{
	VALUEINFO *vi;

	vi = mem_alloc(sizeof(VALUEINFO));
	if(vi == NULL){
		return NULL;
	}
	ZeroMemory(vi, sizeof(VALUEINFO));

	vi->v = mem_alloc(sizeof(VALUE));
	if(vi->v == NULL){
		mem_free(&vi);
		return NULL;
	}
	ZeroMemory(vi->v, sizeof(VALUE));

	vi->v->vi = vi;
	return vi;
}


/******************************************************************************

	Index2Hash

	CfbNXŃnbVl擾

******************************************************************************/

static VALUEINFO *Index2Hash(EXECINFO *ei, VALUEINFO *pvi, int index)
{
	VALUEINFO *vi = pvi;
	VALUEINFO *kvi;
	int i = 0;

	if(index < 0){
		//sȃCfbNX
		Error(ei, ERR_INDEX, ei->Err, FALSE);
		return NULL;
	}

	if(vi->vi == NULL){
		// 0 Ԗڂ̒ǉ
		vi = vi->vi = AllocValue();
		if(vi == NULL){
			Error(ei, ERR_ALLOC, ei->Err, FALSE);
			return NULL;
		}
		vi->v->Global = pvi->v->Global;
		i = 1;
	}else{
		//ʒu
		for(vi = vi->vi; vi != NULL; vi = vi->next, i++){
			if(i == index){
				return vi;
			}
			kvi = vi;
		}
		vi = kvi;
	}
	//
	for(; i <= index; i++){
		vi = vi->next = AllocValue();
		if(vi == NULL){
			Error(ei, ERR_ALLOC, ei->Err, FALSE);
			return NULL;
		}
		vi->v->Global = pvi->v->Global;
	}
	return vi;
}


/******************************************************************************

	GetHashValue

	nbVl̎擾

******************************************************************************/

VALUEINFO *GetHashValue(EXECINFO *ei, VALUEINFO *pvi, VALUE *keyv)
{
	VALUEINFO *kvi;
	VALUEINFO *vi = pvi;
	char *key;
	int name_hash;
	int i = 0;

	if(keyv->Type == TYPE_INTEGER){
		//CfbNXŎQ
		return Index2Hash(ei, pvi, keyv->u.iValue);
	}

	if(*keyv->u.Value == '\0'){
		return pvi;
	}

	//L[ŎQ
	key = keyv->u.Value;
	name_hash = str2hash(key);
	
	if(vi->vi == NULL){
		//VKǉ
		vi = vi->vi = AllocValue();
		if(vi == NULL){
			Error(ei, ERR_ALLOC, ei->Err, FALSE);
			return NULL;
		}
		vi->Name = alloc_copy(key);
		vi->name_hash = name_hash;
		vi->v->Global = pvi->v->Global;
		return vi;
	}

	//
	for(vi = vi->vi; vi != NULL; vi = vi->next){
		if(vi->Name != NULL &&
			name_hash == vi->name_hash && tstrcmp(key, vi->Name) == 0){
			break;
		}
		kvi = vi;
	}
	if(vi == NULL){
		//L[ǉ
		vi = kvi->next = AllocValue();
		if(vi == NULL){
			Error(ei, ERR_ALLOC, ei->Err, FALSE);
			return NULL;
		}
		vi->Name = alloc_copy(key);
		vi->name_hash = name_hash;
		vi->v->Global = pvi->v->Global;
	}
	return vi;
}


/******************************************************************************

	Byte2Hash

	nbVɐݒ

******************************************************************************/

BOOL Byte2Hash(EXECINFO *ei, VALUEINFO *pvi, char *buf, int len)
{
	VALUEINFO *vi = pvi;
	int i = 0;

	if(pvi->v->Global == TRUE) GlobalWait(ei->sci->sci_top);

	if(vi->vi == NULL){
		vi = vi->vi = AllocValue();
		vi->v->Global = pvi->v->Global;
	}else{
		vi = vi->vi;
	}
	vi->v->u.iValue = (unsigned char)*(buf + i);
	vi->v->Type = TYPE_INTEGER;

	for(i = 1; i < len; i++){
		if(vi->next == NULL){
			vi = vi->next = AllocValue();
			vi->v->Global = pvi->v->Global;
		}else{
			vi = vi->next;
		}
		vi->v->u.iValue = (unsigned char)*(buf + i);
		vi->v->Type = TYPE_INTEGER;
	}
	if(pvi->v->Global == TRUE) GlobalUnWait(ei->sci->sci_top);
	return TRUE;
}


/******************************************************************************

	Hash2Byte

	nbV𕶎ɐݒ

******************************************************************************/

int Hash2Byte(VALUEINFO *pvi, char **ret)
{
	VALUEINFO *vi;
	char *p;
	int len;

	if(pvi == NULL){
		*ret = (char *)alloc_copy("");
		return 0;
	}
	for(len = 0, vi = pvi->vi; vi != NULL; vi = vi->next, len++);
	p = *ret = mem_alloc(len);
	if(p == NULL){
		*ret = (char *)alloc_copy("");
		return 0;
	}

	for(vi = pvi->vi; vi != NULL; vi = vi->next){
		*(p++) = (char)((vi->v->Type == TYPE_INTEGER) ? vi->v->u.iValue : *vi->v->u.Value);
	}
	return len;
}


/******************************************************************************

	GetValueString

	̎擾

******************************************************************************/

char *GetValueString(VALUE *v)
{
	char *buf;

	buf = ((v->Type == TYPE_INTEGER) ? i2a(v->u.iValue) : alloc_copy(v->u.Value));
	if(buf == NULL){
		buf = alloc_copy("");
	}
	return buf;
}


/******************************************************************************

	GetValueString

	̎擾 (TCHAR)

******************************************************************************/

TCHAR *TGetValueString(VALUE *v)
{
	TCHAR *tbuf;
	char *buf;

	buf = ((v->Type == TYPE_INTEGER) ? i2a(v->u.iValue) : alloc_copy(v->u.Value));
	if(buf == NULL){
		tbuf = char2tchar("");
	}else{
		tbuf = char2tchar(buf);
		mem_free(&buf);
	}
	return tbuf;
}


/******************************************************************************

	GetValueInt

	l̎擾

******************************************************************************/

long GetValueInt(VALUE *v)
{
	return ((v->Type == TYPE_INTEGER) ? v->u.iValue : atol(v->u.Value));
}


/******************************************************************************

	SetValue

	l̐ݒ

******************************************************************************/

void SetValue(EXECINFO *ei, VALUE *to_v, VALUE *from_v)
{
	char *p = NULL;

	//bN
	if(to_v->Global == TRUE && ei != NULL) GlobalWait(ei->sci->sci_top);
	if(to_v->Type == TYPE_STRING) p = to_v->u.Value;

	if(from_v->Type == TYPE_INTEGER){
		//l
		to_v->u.iValue = from_v->u.iValue;
	}else{
		//
		to_v->u.Value = alloc_copy(from_v->u.Value);
	}
	to_v->Type = from_v->Type;

	if(p != NULL) mem_free(&p);
	if(to_v->Global == TRUE && ei != NULL) GlobalUnWait(ei->sci->sci_top);
}


/******************************************************************************

	GetLocalVariable

	[Jϐ擾

******************************************************************************/

VALUEINFO *GetLocalVariable(EXECINFO *ei, char *name)
{
	VALUEINFO *vi = NULL;

	//ϐ̌
	for(; ei != NULL; ei = ei->parent){
		vi = FindValueInfo(ei->vi, name);
		if(vi != NULL){
			break;
		}
	}
	return vi;
}


/******************************************************************************

	SetLocalVariable

	[Jϐݒ

******************************************************************************/

BOOL SetLocalVariable(EXECINFO *ei, char *name, VALUE *v)
{
	VALUEINFO *vi;

	//ϐ̐ݒ
	vi = GetLocalVariable(ei, name);
	if(vi != NULL){
		SetValue(ei, vi->v, v);
		return TRUE;
	}
	return ((AddValueInfo(&(ei->vi), name, v) == NULL) ? FALSE : TRUE);
}


/******************************************************************************

	GetGlobalVariable

	O[oϐ擾

******************************************************************************/

VALUEINFO *GetGlobalVariable(EXECINFO *ei, char *rname)
{
	SCRIPTINFO *csci = ei->sci;
	char *name;

	if(ei->sci == NULL) return NULL;

	//ΏۃXNvǧ
	for(name = rname; *name != '\0' && *name != SCRIPT_SEP; name++);
	if(*name != SCRIPT_SEP){
		name = rname;
	}else{
		name++;
		csci = GetScriptInfo(ei->sci, rname);
		if(csci == NULL){
			Error(ei, ERR_SCRIPT, rname, FALSE);
			return NULL;
		}
	}

	//ϐ̌
	return FindValueInfo(csci->vi, name);
}


/******************************************************************************

	SetGlobalVariable

	O[oϐݒ

******************************************************************************/

BOOL SetGlobalVariable(EXECINFO *ei, char *rname, VALUE *v)
{
	SCRIPTINFO *csci = ei->sci;
	VALUEINFO *vi;
	char *name;
	BOOL ret;

	//ΏۃXNvǧ
	for(name = rname; *name != '\0' && *name != SCRIPT_SEP; name++);
	if(*name != SCRIPT_SEP){
		name = rname;
	}else{
		name++;
		csci = GetScriptInfo(ei->sci, rname);
		if(csci == NULL){
			Error(ei, ERR_SCRIPT, rname, FALSE);
			return FALSE;
		}
	}

	//ϐ̐ݒ
	vi = FindValueInfo(csci->vi, name);
	if(vi != NULL){
		vi->v->Global = TRUE;
		SetValue(ei, vi->v, v);
		ret = TRUE;
	}else{
		vi = AddValueInfo(&(csci->vi), name, v);
		if(vi == NULL){
			Error(ei, ERR_ALLOC, ei->Err, FALSE);
			ret = FALSE;
		}else{
			vi->v->Global = TRUE;
		}
	}
	return ret;
}


/******************************************************************************

	DeclVariable

	ϐ`

******************************************************************************/

static VALUEINFO *DeclVariable(EXECINFO *ei, char *name, char *err)
{
	char *str;
	VALUE v;

	str = (*name == ARGUMENT_ADDRESS) ? name + 1 : name;

	if(FindValueInfo(ei->vi, str) != NULL){
		//ϐ̏d`
		char *p = alloc_join(str, ERR_DECLARE);
		Error(ei, p, err, FALSE);
		mem_free(&p);
		return NULL;
	}
	ZeroMemory(&v, sizeof(VALUE));
	return AddValueInfo(&(ei->vi), str, &v);
}


/******************************************************************************

	GetVariable

	ϐ擾

******************************************************************************/

VALUEINFO *GetVariable(EXECINFO *ei, char *name)
{
	VALUEINFO *vi;

	//[Jϐ
	vi = GetLocalVariable(ei, name);
	if(vi == NULL){
		//O[oϐ
		vi = GetGlobalVariable(ei, name);
	}
	return vi;
}


/******************************************************************************

	FindValueInfo

	ϐ̌

******************************************************************************/

static VALUEINFO *FindValueInfo(VALUEINFO *vi, char *name)
{
	int name_hash;

	name_hash = str2hash(name);
	for(; vi != NULL; vi = vi->next){
		if(name_hash == vi->name_hash && tstrcmp(name, vi->Name) == 0){
			return vi;
		}
	}
	return NULL;
}


/******************************************************************************

	AddValueInfo

	ϐ̒ǉ

******************************************************************************/

static VALUEINFO *AddValueInfo(VALUEINFO **vi_root, char *name, VALUE *v)
{
	VALUEINFO *vi;

	if(*vi_root == NULL){
		vi = *vi_root = AllocValue();
	}else{
		for(vi = *vi_root; vi->next != NULL; vi = vi->next);
		vi = vi->next = AllocValue();
	}
	if(vi == NULL){
		return NULL;
	}
	vi->Name = alloc_copy(name);
	if(v == NULL){
		vi->v->u.iValue = 0;
		vi->v->Type = TYPE_INTEGER;
	}else{
		SetValue(NULL, vi->v, v);
	}
	vi->name_hash = str2hash(vi->Name);
	return vi;
}


/******************************************************************************

	CopyValueList

	ϐXg̃Rs[

******************************************************************************/

VALUEINFO *CopyValueList(VALUEINFO *From)
{
	VALUEINFO *To, Top;
	char *p;

	To = &Top;
	To->next = NULL;
	for(; From != NULL; From = From->next){
		To = To->next = AllocValue();
		if(To == NULL){
			return Top.next;
		}
		if(From->Name != NULL){
			To->Name = alloc_copy(From->Name);
			To->name_hash = From->name_hash;
		}
		if(From->v->Type == TYPE_STRING){
			p = alloc_copy(From->v->u.Value);
			To->v->u.Value = p;
		}else{
			To->v->u.iValue = From->v->u.iValue;
		}
		To->v->Type = From->v->Type;
		To->vi = CopyValueList(From->vi);
	}
	return Top.next;
}


/******************************************************************************

	VariableFromString

	񂩂ϐ擾 (%`%)

******************************************************************************/

VALUEINFO *VariableFromString(EXECINFO *ei, char *param)
{
	VALUEINFO *stack = NULL;
	TOKEN *tk;
	char *p, *r;

	for(r = param + 1; *r != '\0' && *r != VARIABLE_CHAR; r++);
	p = alloc_copy_n(param + 1, r - param);
	if(p == NULL){
		return NULL;
	}

	//WJ
	tk = ParseVariable(ei, p);
	if(tk == NULL){
		Error(ei, ERR_SENTENCE, param, FALSE);
		mem_free(&p);
		return NULL;
	}

	if(ExecSentense(ei, tk, NULL, &stack, NULL) == RET_ERROR){
		FreeToken(tk);
		mem_free(&p);
		return NULL;
	}
	if(stack != NULL){
		FreeValueList(stack->next);
		stack->next = NULL;
	}
	FreeToken(tk);
	mem_free(&p);
	return stack;
}


/******************************************************************************

	GetStringConvLen

	ϊTCY擾

******************************************************************************/

static int GetStringConvLen(EXECINFO *ei, char *fmt, VALUEINFO *vi)
{
	char *s;
	int ret = 0;

	while(*fmt != '\0'){
		if(IsDBCSLeadByte((BYTE)*fmt) == TRUE && *(fmt + 1) != '\0'){
			//2oCgR[hXLbv
			fmt += 2;
			ret += 2;
			continue;
		}
		if(*fmt == VARIABLE_CHAR){
			if(*(fmt + 1) == VARIABLE_CHAR){
				fmt += 2;
				ret++;
				continue;
			}
			vi = vi->next = VariableFromString(ei, fmt);
			if(vi == NULL){
				return -1;
			}
			if(vi->v->Type == TYPE_INTEGER){
				s = i2a(vi->v->u.iValue);
			}else{
				s = vi->v->u.Value;
			}
			if(s != NULL){
				ret += tstrlen(s);
			}
			if(vi->v->Type == TYPE_INTEGER){
				mem_free(&s);
			}
			for(fmt++; *fmt != '\0' && *fmt != VARIABLE_CHAR; *fmt++);
			if(*fmt == '\0'){
				break;
			}
		}else{
			ret++;
		}
		fmt++;
	}
	return ret;
}


/******************************************************************************

	StringConv

	ϊ

******************************************************************************/

static char *StringConv(EXECINFO *ei, char *fmt)
{
	VALUEINFO *vi, rvi;
	char *ret;
	char *p, *pp, *r, *s;
	int len;

	p = pp = alloc_copy(fmt);
	conv_ctrl(p);

	rvi.next = NULL;
	len = GetStringConvLen(ei, p, &rvi);
	if(len < 0){
		mem_free(&pp);
		FreeValueList(rvi.next);
		return NULL;
	}
	if(rvi.next == NULL){
		//ϐ
		for(p = pp; *p != '\0'; p++){
			if(*p != '%'){
				continue;
			}
			tstrcpy(p, p + 1);
			p++;
			if(*p == '\0'){
				break;
			}
		}
		return pp;
	}
	vi = rvi.next;

	r = ret = (char *)mem_alloc(sizeof(char) * (len + 1));
	if(ret == NULL){
		Error(ei, ERR_ALLOC, ei->Err, FALSE);
		mem_free(&pp);
		FreeValueList(rvi.next);
		return NULL;
	}

	while(*p != '\0'){
		if(IsDBCSLeadByte((BYTE)*p) == TRUE && *(p + 1) != '\0'){
			//2oCgR[hXLbv
			*(r++) = *(p++);
			*(r++) = *(p++);
			continue;
		}
		if(*p == VARIABLE_CHAR){
			if(*(p + 1) == VARIABLE_CHAR){
				*(r++) = *(p++);
				p++;
				continue;
			}
			//ϐ̕ϊ
			if(vi != NULL){
				if(vi->v->Type == TYPE_INTEGER){
					s = i2a(vi->v->u.iValue);
				}else{
					s = vi->v->u.Value;
				}
				if(s != NULL){
					r = str_cpy(r, s);
				}
				if(vi->v->Type == TYPE_INTEGER){
					mem_free(&s);
				}
				vi = vi->next;
			}
			for(p++; *p != '\0' && *p != VARIABLE_CHAR; *p++);

		}else{
			*(r++) = *p;
		}
		p++;
	}
	*r = '\0';
	mem_free(&pp);
	FreeValueList(rvi.next);
	return ret;
}


/******************************************************************************

	FindCase

	case ڂ̌

******************************************************************************/

static TOKEN *FindCase(EXECINFO *ei, TOKEN *cu_tk, VALUEINFO *v1)
{
	VALUEINFO *vi, *v2;
	TOKEN *tmp_tk;
	int cp;

	//case ڂ̌
	for(tmp_tk = cu_tk->target; tmp_tk != NULL; tmp_tk = tmp_tk->next){
		if(tmp_tk->Type != SYM_CASE){
			continue;
		}
		v2 = NULL;
		ei->to_tk = SYM_CASEEND;
		if(ExecSentense(ei, tmp_tk->next, NULL, &v2, NULL) == -1){
			return NULL;
		}
		ei->to_tk = 0;

		//X^bN̓eo
		if(v2 == NULL){
			return NULL;
		}

		//r
		vi = CalcValue(ei, v2, v1, SYM_EQEQ);
		FreeValue(v2);
		if(vi == NULL){
			return NULL;
		}
		cp = vi->v->u.iValue;
		FreeValue(vi);
		if(cp != 0){
			//}b`
			break;
		}
	}
	//default ̌
	if(tmp_tk == NULL){
		for(tmp_tk = cu_tk->target; tmp_tk != NULL &&
			tmp_tk->Type != SYM_DEFAULT; tmp_tk = tmp_tk->next);
	}
	return tmp_tk;
}


/******************************************************************************

	UnaryCalcValue

	PZq̌vZs

******************************************************************************/

static VALUEINFO *UnaryCalcValue(EXECINFO *ei, VALUEINFO *vi, int c)
{
	VALUEINFO *vret;
	int i;

	i = GetValueInt(vi->v);
	switch(c)
	{
	case SYM_NOT:
		i = !i;
		break;

	case SYM_BITNOT:
		i = ~i;
		break;

	case SYM_PLUS:
		i = +i;
		break;

	case SYM_MINS:
		i = -i;
		break;

	case SYM_INC:
		vi->v->u.iValue = ++i;
		vi->v->Type = TYPE_INTEGER;
		break;

	case SYM_DEC:
		vi->v->u.iValue = --i;
		vi->v->Type = TYPE_INTEGER;
		break;

	case SYM_BINC:
	case SYM_BDEC:
		if(vi == vi->v->vi){
			return vi;
		}
		vret = AllocValue();
		if(vret == NULL){
			Error(ei, ERR_ALLOC, ei->Err, FALSE);
			return NULL;
		}
		mem_free(&vret->v);
		vret->v = vi->v;

		switch(c)
		{
		case SYM_BINC:
			vret->next = ei->inc_vi;
			ei->inc_vi = vret;
			break;
		case SYM_BDEC:
			vret->next = ei->dec_vi;
			ei->dec_vi = vret;
			break;
		}
		return vi;
	}
	vret = AllocValue();
	if(vret == NULL){
		Error(ei, ERR_ALLOC, ei->Err, FALSE);
		return NULL;
	}
	vret->v->u.iValue = i;
	vret->v->Type = TYPE_INTEGER;
	return vret;
}


/******************************************************************************

	IntegerCalcValue

	l^̌vZs

******************************************************************************/

static VALUEINFO *IntegerCalcValue(EXECINFO *ei, VALUEINFO *v1, VALUEINFO *v2, int c)
{
	VALUEINFO *vret;
	long i = 0, j;

	i = v1->v->u.iValue;
	j = v2->v->u.iValue;
	switch(c)
	{
	case SYM_DIV:
	case SYM_MOD:
		if(j == 0){
			Error(ei, ERR_DIVZERO, ei->Err, FALSE);
			return NULL;
		}
		if(c == SYM_DIV){
			i /= j;
		}else{
			i %= j;
		}
		break;

	case SYM_MULTI: i *= j; break;
	case SYM_ADD: i += j; break;
	case SYM_SUB: i -= j; break;
	case SYM_AND: i &= j; break;
	case SYM_OR: i |= j; break;
	case SYM_XOR: i ^= j; break;
	case SYM_CPAND: i = i && j; break;
	case SYM_CPOR: i = i || j; break;
	case SYM_REGEQ:
	case SYM_EQEQ: i = i == j; break;
	case SYM_NTEQ: i = i != j; break;
	case SYM_LEFTEQ: i = i <= j; break;
	case SYM_LEFT: i = i < j; break;
	case SYM_RIGHTEQ: i = i >= j; break;
	case SYM_RIGHT: i = i > j; break;
	case SYM_LEFTSHIFT: i = (unsigned long)i << j; break;
	case SYM_RIGHTSHIFT: i = (unsigned long)i >> j; break;

	default:
		Error(ei, ERR_OPERATOR, ei->Err, FALSE);
		return NULL;
	}

	vret = AllocValue();
	if(vret == NULL){
		Error(ei, ERR_ALLOC, ei->Err, FALSE);
		return NULL;
	}
	vret->v->u.iValue = i;
	vret->v->Type = TYPE_INTEGER;
	return vret;
}


/******************************************************************************

	StringCalcValue

	^̌vZs

******************************************************************************/

static VALUEINFO *StringCalcValue(EXECINFO *ei, VALUEINFO *v1, VALUEINFO *v2, int c)
{
	VALUEINFO *vret;
	rege_nfa_t *nfa;
	char *p, *s;
	char *p1, *p2;
	char *f1 = NULL, *f2 = NULL;
	long i = 0, j;

	if(v1->v->Type == TYPE_INTEGER){
		f1 = p1 = i2a(v1->v->u.iValue);
	}else{
		p1 = v1->v->u.Value;
	}
	if(v2->v->Type == TYPE_INTEGER){
		f2 = p2 = i2a(v2->v->u.iValue);
	}else{
		p2 = v2->v->u.Value;
	}

	switch(c)
	{
	case SYM_ADD:
		//񌋍
		p = mem_alloc(sizeof(char) * (tstrlen(p1) + tstrlen(p2) + 1));
		if(p == NULL){
			Error(ei, ERR_ALLOC, ei->Err, FALSE);
			mem_free(&f1);
			mem_free(&f2);
			return NULL;
		}
		s = str_cpy(p, p1);
		s = str_cpy(s, p2);

		vret = AllocValue();
		if(vret == NULL){
			Error(ei, ERR_ALLOC, ei->Err, FALSE);
			mem_free(&f1);
			mem_free(&f2);
			mem_free(&p);
			return NULL;
		}
		vret->v->u.Value = p;
		vret->v->Type = TYPE_STRING;
		mem_free(&f1);
		mem_free(&f2);
		return vret;

	case SYM_CPAND:
		i = (v1->v->Type == TYPE_INTEGER) ? v1->v->u.iValue : *v1->v->u.Value;
		j = (v2->v->Type == TYPE_INTEGER) ? v2->v->u.iValue : *v2->v->u.Value;
		i = i && j;
		break;

	case SYM_CPOR:
		i = (v1->v->Type == TYPE_INTEGER) ? v1->v->u.iValue : *v1->v->u.Value;
		j = (v2->v->Type == TYPE_INTEGER) ? v2->v->u.iValue : *v2->v->u.Value;
		i = i || j;
		break;

	case SYM_EQEQ:
		i = (*p1 == *p2 && tstrcmp(p1, p2) == 0) ? 1 : 0;
		break;

	case SYM_REGEQ:
		if(v2->nfa != NULL){
			i = (reg_match_value(NULL, v2->nfa, p1) == TRUE) ? 1 : 0;
		}else{
			//K\̃RpC
			nfa = rege_compile(p2);
			i = (reg_match_value(NULL, nfa, p1) == TRUE) ? 1 : 0;
			free_nfa(nfa);
		}
		break;

	case SYM_NTEQ:
		i = (*p1 == *p2 && tstrcmp(p1, p2) == 0) ? 0 : 1;
		break;

	default:
		Error(ei, ERR_STROPERATOR, ei->Err, FALSE);
		mem_free(&f1);
		mem_free(&f2);
		return NULL;
	}
	mem_free(&f1);
	mem_free(&f2);

	vret = AllocValue();
	if(vret == NULL){
		Error(ei, ERR_ALLOC, ei->Err, FALSE);
		return NULL;
	}
	vret->v->u.iValue = i;
	vret->v->Type = TYPE_INTEGER;
	return vret;
}


/******************************************************************************

	HashCalcValue

	AzžvZs

******************************************************************************/

static VALUEINFO *HashCalcValue(EXECINFO *ei, VALUEINFO *v1, VALUEINFO *v2, int c)
{
	VALUEINFO *vret;
	VALUEINFO *vi;
	long i = 0;

	switch(c)
	{
	case SYM_ADD:
		//Azz̘A
		vret = AllocValue();
		if(vret == NULL){
			Error(ei, ERR_ALLOC, ei->Err, FALSE);
			return NULL;
		}
		vret->vi = CopyValueList(v1->vi);
		if(vret->vi != NULL){
			for(vi = vret->vi; vi->next != NULL; vi = vi->next);
			//A
			vi->next = CopyValueList(v2->vi);
		}else{
			vret->vi = CopyValueList(v2->vi);
		}
		vret->Hash = TRUE;
		return vret;

	case SYM_EQEQ:
	case SYM_NTEQ:
		//Azz̓er
		for(v1 = v1->vi, v2 = v2->vi; v1 != NULL && v2 != NULL; v1 = v1->next, v2 = v2->next){
			vi = CalcValue(ei, v1, v2, SYM_EQEQ);
			i = vi->v->u.iValue;
			FreeValue(vi);
			if(i == 0){
				break;
			}
		}
		i = (v1 == NULL && v2 == NULL) ? 1 : 0;
		if(c == SYM_NTEQ){
			i = !i;
		}
		break;

	default:
		Error(ei, ERR_HASHOPERATOR, ei->Err, FALSE);
		return NULL;
	}

	vret = AllocValue();
	if(vret == NULL){
		Error(ei, ERR_ALLOC, ei->Err, FALSE);
		return NULL;
	}
	vret->v->u.iValue = i;
	vret->v->Type = TYPE_INTEGER;
	return vret;
}


/******************************************************************************

	CalcValue

	񍀉Zq̌vZs

******************************************************************************/

static VALUEINFO *CalcValue(EXECINFO *ei, VALUEINFO *v1, VALUEINFO *v2, int c)
{
	if(v1->Hash == TRUE || v2->Hash == TRUE){
		//AzžvZ
		return HashCalcValue(ei, v1, v2, c);

	}else if(v1->v->Type == TYPE_INTEGER && v2->v->Type == TYPE_INTEGER){
		//l^̌vZ
		return IntegerCalcValue(ei, v1, v2, c);
	}
	//^̌vZ
	return StringCalcValue(ei, v1, v2, c);
}


/******************************************************************************

	ExecSentense

	͖؂̎ss

******************************************************************************/

int ExecSentense(EXECINFO *ei, TOKEN *cu_tk, VALUEINFO *retvi, VALUEINFO **retstack, int *breakcnt)
{
	EXECINFO cei;
	VALUEINFO *vi, *v1, *v2;
	VALUEINFO *stack = NULL;
	TOKEN *tmp_tk;
	char *r;
	int RetSt = RET_SUCCESS;
	int cp;

	//XNvgICxg̃`FbN
	if(CheckEndEvent() == TRUE){
		return RET_ERROR;
	}

	while(cu_tk != NULL && cu_tk->Type != ei->to_tk){
		ei->Err = cu_tk->Err;

		switch(cu_tk->Type)
		{
		case SYM_BOPEN:
			ZeroMemory(&cei, sizeof(EXECINFO));
			cei.sci = ei->sci;
			cei.lParam = ei->lParam;
			cei.parent = ei;
			vi = NULL;
			RetSt = ExecSentense(&cei, cu_tk->target, retvi, &vi, breakcnt);
			FreeExecInfo(&cei);
			if(vi == NULL){
				break;
			}
			//s̃X^bN̓eAzzɐݒ
			v1 = AllocValue();
			if(v1 == NULL){
				Error(ei, ERR_ALLOC, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			v1->vi = vi;
			v1->Hash = TRUE;
			v1->next = stack;
			stack = v1;
			break;

		case SYM_LINEEND:
			//X^bN
			FreeValueList(stack);
			stack = NULL;
			break;

		case SYM_WORDEND:
			if(stack == NULL){
				//X^bNɋ̒lǉ
				stack = AllocValue();
				if(stack == NULL){
					Error(ei, ERR_ALLOC, ei->Err, FALSE);
					RetSt = RET_ERROR;
					break;
				}
			}
			if(cu_tk->next != NULL && cu_tk->next->Type == SYM_WORDEND){
				//X^bNɋ̒lǉ
				vi = AllocValue();
				if(vi == NULL){
					Error(ei, ERR_ALLOC, ei->Err, FALSE);
					RetSt = RET_ERROR;
					break;
				}
				vi->next = stack;
				stack = vi;
			}
			break;

		case SYM_JUMP:
			//ړ
			cu_tk = cu_tk->link;
			break;

		case SYM_RETURN:
		case SYM_EXIT:
			if(cu_tk->Type == SYM_EXIT){
				RetSt = RET_EXIT;
			}else{
				RetSt = RET_RETURN;
			}
			if(retvi == NULL){
				break;
			}
			if(stack == NULL){
				retvi->v->u.iValue = 0;
				retvi->v->Type = TYPE_INTEGER;
				break;
			}
			//߂l
			vi = stack;
			stack = stack->next;
			SetValue(ei, retvi->v, vi->v);
			retvi->vi = CopyValueList(vi->vi);
			retvi->Hash = vi->Hash;
			FreeValue(vi);
			break;

		case SYM_BREAK:
			//f
			RetSt = RET_BREAK;
			if(breakcnt == NULL){
				break;
			}
			if(stack == NULL){
				*breakcnt = 0;
				break;
			}
			//break
			vi = stack;
			stack = stack->next;
			*breakcnt = GetValueInt(vi->v);
			if(*breakcnt < 0){
				*breakcnt = 0;
			}
			FreeValue(vi);
			break;

		case SYM_CONTINUE:
			//p
			RetSt = RET_CONTINUE;
			break;

		case SYM_CASE:
		case SYM_DEFAULT:
			for(; cu_tk->next != NULL && cu_tk->Type != SYM_CASEEND; cu_tk = cu_tk->next);
			break;

		case SYM_CMP:
			//r
			if(stack == NULL){
				cu_tk = cu_tk->next;
				break;
			}
			//r
			vi = stack;
			stack = stack->next;
			cp = (vi->v->Type == TYPE_INTEGER) ? vi->v->u.iValue : *vi->v->u.Value;
			FreeValue(vi);
			if(cp != 0){
				//^
				cu_tk = cu_tk->next;
			}
			break;

		case SYM_ELSE:
		case SYM_CMPEND:
			break;

		case SYM_SWITCH:
			//򕪊
			if(stack == NULL){
				Error(ei, ERR_SENTENCE, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			vi = stack;
			stack = stack->next;

			//case ڂ̌
			tmp_tk = FindCase(ei, cu_tk, vi);
			FreeValue(vi);
			if(tmp_tk == NULL){
				//}b` case ڂ
				break;
			}
			//̎s
			ZeroMemory(&cei, sizeof(EXECINFO));
			cei.sci = ei->sci;
			cei.lParam = ei->lParam;
			cei.parent = ei;
			RetSt = ExecSentense(&cei, tmp_tk, retvi, NULL, breakcnt);
			FreeExecInfo(&cei);
			if(RetSt == RET_BREAK && breakcnt != NULL && *breakcnt > 0){
				(*breakcnt)--;
				break;
			}
			if(RetSt != RET_SUCCESS && RetSt != RET_BREAK){
				break;
			}
			RetSt = RET_SUCCESS;
			break;

		case SYM_LOOP:
			//JԂ
			if(stack != NULL){
				//r
				vi = stack;
				stack = stack->next;

				cp = (vi->v->Type == TYPE_INTEGER) ? vi->v->u.iValue : *vi->v->u.Value;
				FreeValue(vi);
				if(cp == 0){
					break;
				}
			}
			//[vΏۏ
			RetSt = ExecSentense(ei, cu_tk->target, retvi, NULL, breakcnt);
			if(RetSt != RET_SUCCESS && RetSt != RET_BREAK && RetSt != RET_CONTINUE){
				break;
			}
			if(RetSt == RET_BREAK){
				//[v𒆒f
				if(breakcnt != NULL && *breakcnt > 0){
					(*breakcnt)--;
					break;
				}
				for(; cu_tk->next != NULL && cu_tk->Type != SYM_LOOPEND; cu_tk = cu_tk->next);
				RetSt = RET_SUCCESS;
				break;
			}
			//p
			cu_tk = cu_tk->next;
			RetSt = RET_SUCCESS;
			Sleep(0);
			break;

		case SYM_LOOPEND:
		case SYM_LOOPSTART:
			break;

		case SYM_ARGSTART:
			//̊Jnʒu
			vi = AllocValue();
			if(vi == NULL){
				Error(ei, ERR_ALLOC, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			mem_free(&vi->v);
			vi->next = stack;
			stack = vi;
			break;

		case SYM_FUNC:
			//֐Ăяo
			if(stack == NULL){
				Error(ei, ERR_SENTENCE, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			//̎擾
			v1 = NULL;
			while(stack != NULL && stack->v != NULL){
				vi = stack;
				stack = stack->next;
				vi->next = v1;
				v1 = vi;
			}
			//Jnʒủ
			vi = stack;
			stack = stack->next;
			FreeValue(vi);

			if(*cu_tk->buf == THREAD_EXEC){
				//VXbhŎs
				v2 = AllocValue();
				if(v2 == NULL){
					Error(ei, ERR_ALLOC, ei->Err, FALSE);
					RetSt = RET_ERROR;
					break;
				}
				v2->v->u.iValue = CreateScriptThread(ei->sci, cu_tk, cu_tk->buf + 1, v1, FALSE, ei->lParam);
			}else{
				v2 = ExecFunction(ei, cu_tk, cu_tk->buf, v1, FALSE);
			}
			FreeValueList(v1);
			if(v2 == NULL || (int)v2 == -1){
#ifndef NO_CALLLIST
				EXECINFO *pei;
				//֐Ăяȍo
				for(pei = ei; pei->parent != NULL; pei = pei->parent);
				if(tstrcmp(pei->Name, cu_tk->buf) != 0){
					Error(ei, NULL, ei->Err, TRUE);
				}
#endif
				RetSt = RET_ERROR;
				break;
			}
			//߂lX^bNɐς
			v2->next = stack;
			stack = v2;
			break;

		case SYM_DECLVARIABLE:
			//ϐ`
			v1 = DeclVariable(ei, cu_tk->buf, ei->Err);
			if(v1 == NULL){
				RetSt = RET_ERROR;
				break;
			}

			vi = AllocValue();
			if(vi == NULL){
				Error(ei, ERR_ALLOC, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			mem_free(&vi->v);
			vi->v = v1->v;
			vi->vi = v1->vi;
			vi->next = stack;
			stack = vi;
			break;

		case SYM_VARIABLE:
			//ϐ擾
			r = (*cu_tk->buf == ARGUMENT_ADDRESS) ? cu_tk->buf + 1 : cu_tk->buf;
			v1 = GetVariable(ei, r);
			if(v1 == NULL){
				if(ei->sci->explicit_val == TRUE){
					//ϐ`
					char *err = alloc_join(r, ERR_NOTDECLARE);
					Error(ei, err, ei->Err, FALSE);
					mem_free(&err);
					RetSt = RET_ERROR;
					break;
				}
				//ϐ`
				v1 = DeclVariable(ei, cu_tk->buf, ei->Err);
				if(v1 == NULL){
					RetSt = RET_ERROR;
					break;
				}
			}
			vi = AllocValue();
			if(vi == NULL){
				Error(ei, ERR_ALLOC, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			if(retstack != NULL){
				vi->Name = alloc_copy(v1->Name);
				vi->name_hash = v1->name_hash;
			}
			mem_free(&vi->v);
			vi->v = v1->v;
			vi->vi = v1->vi;
			vi->next = stack;
			stack = vi;
			break;

		case SYM_HASH:
			//Azz
			if(stack == NULL || stack->next == NULL){
				Error(ei, ERR_SENTENCE, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			v1 = stack;
			stack = stack->next;
			v2 = stack;
			stack = stack->next;

			vi = GetHashValue(ei, v2->v->vi, v1->v);
			if(vi == NULL){
				FreeValue(v1);
				FreeValue(v2);
				RetSt = RET_ERROR;
				break;
			}
			//L[󕶎̏ꍇ̓nbV^ƂĈ
			cp = (v1->v->Type == TYPE_STRING && *v1->v->u.Value == '\0') ? TRUE : FALSE;
			FreeValue(v1);

			v1 = AllocValue();
			if(v1 == NULL){
				FreeValue(v2);
				Error(ei, ERR_ALLOC, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			if(v2->v != NULL && v2->v->vi == v2){
				//萔̏ꍇ̓Rs[
				SetValue(ei, v1->v, vi->v);
				v1->v->vi->vi = CopyValueList(vi->vi);
			}else{
				//ϐ̒l̎Q
				mem_free(&v1->v);
				v1->v = vi->v;
				v1->vi = vi->vi;
			}
			v1->Hash = cp;
			v1->next = stack;
			stack = v1;
			FreeValue(v2);
			break;

		case SYM_CONST:
			//l萔
			vi = AllocValue();
			if(vi == NULL){
				Error(ei, ERR_ALLOC, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			vi->v->u.iValue = cu_tk->i;
			vi->v->Type = TYPE_INTEGER;

			vi->next = stack;
			stack = vi;
			break;

		case SYM_STRING:
			//萔
			vi = AllocValue();
			if(vi == NULL){
				Error(ei, ERR_ALLOC, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			vi->v->u.Value = StringConv(ei, cu_tk->buf);
			if(vi->v->u.Value == NULL){
				RetSt = RET_ERROR;
				break;
			}
			vi->v->Type = TYPE_STRING;
			vi->nfa = cu_tk->nfa;
			vi->next = stack;
			stack = vi;
			break;

		case SYM_ASSIGNSTRING:
			//Ώۂ̕
			vi = AllocValue();
			if(vi == NULL){
				Error(ei, ERR_ALLOC, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			vi->nfa = cu_tk->nfa;
			vi->next = stack;
			stack = vi;
			break;

		case SYM_NOT:
		case SYM_BITNOT:
		case SYM_PLUS:
		case SYM_MINS:
		case SYM_INC:
		case SYM_DEC:
		case SYM_BINC:
		case SYM_BDEC:
			//PZq
			if(stack == NULL){
				Error(ei, ERR_SENTENCE, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			vi = stack;
			stack = stack->next;

			v1 = UnaryCalcValue(ei, vi, cu_tk->Type);
			if(v1 != vi){
				FreeValue(vi);
			}
			if(v1 == NULL){
				RetSt = RET_ERROR;
				break;
			}
			v1->next = stack;
			stack = v1;
			break;

		case SYM_POSTFIX:
			//u
			while(ei->inc_vi != NULL){
				vi = ei->inc_vi;
				ei->inc_vi = ei->inc_vi->next;
				vi->v->u.iValue = GetValueInt(vi->v) + 1;
				vi->v->Type = TYPE_INTEGER;
				FreeValue(vi);
			}
			while(ei->dec_vi != NULL){
				vi = ei->dec_vi;
				ei->dec_vi = ei->dec_vi->next;
				vi->v->u.iValue = GetValueInt(vi->v) - 1;
				vi->v->Type = TYPE_INTEGER;
				FreeValue(vi);
			}
			break;

		case SYM_EQ:
			//Zq
			if(stack == NULL || stack->next == NULL){
				Error(ei, ERR_SENTENCE, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			//E
			v1 = stack;
			stack = stack->next;
			//
			v2 = stack;
			stack = stack->next;

			if(v2->nfa != NULL){
				//K\œWJĕϐɐݒ
				r = (v1->v->Type == TYPE_INTEGER) ? i2a(v1->v->u.iValue) : v1->v->u.Value;
				reg_match_value(ei, v2->nfa, r);
				if(v1->v->Type == TYPE_INTEGER){
					mem_free(&r);
				}
			}else if(v1->Hash == TRUE || v2->Hash == TRUE){
				//Azz̃Rs[
				FreeValueList(v2->v->vi->vi);
				v2->v->vi->vi = CopyValueList(v1->vi);
			}else{
				//ϐɒlݒ
				SetValue(ei, v2->v, v1->v);
			}
			if(retstack != NULL){
				mem_free(&v1->Name);
				v1->Name = alloc_copy(v2->Name);
				v1->name_hash = v2->name_hash;
			}
			//lX^bNɐς
			v1->next = stack;
			stack = v1;
			FreeValue(v2);
			break;

		default:
			//񍀉Zq
			if(stack == NULL || stack->next == NULL){
				Error(ei, ERR_SENTENCE, ei->Err, FALSE);
				RetSt = RET_ERROR;
				break;
			}
			v1 = stack;
			stack = stack->next;
			v2 = stack;
			stack = stack->next;

			vi = CalcValue(ei, v2, v1, cu_tk->Type);
			if(vi == NULL){
				RetSt = RET_ERROR;
				break;
			}
			FreeValue(v1);
			FreeValue(v2);

			//ZʂX^bNɐς
			vi->next = stack;
			stack = vi;
			break;
		}
		if(RetSt != RET_SUCCESS){
			break;
		}
		cu_tk = cu_tk->next;
	}

	if(retstack != NULL){
		//X^bNԂ
		*retstack = NULL;
		while(stack != NULL){
			vi = stack;
			stack = stack->next;
			vi->next = *retstack;
			*retstack = vi;
		}
	}else{
		//X^bN
		FreeValueList(stack);
	}
	return RetSt;
}


/******************************************************************************

	ExpandArgument

	̓WJ

******************************************************************************/

static TOKEN *ExpandArgument(EXECINFO *ei, TOKEN *tk, VALUEINFO *param)
{
	VALUEINFO *vi;
	char *str;
	BOOL eq = FALSE;

	while(tk != NULL && tk->Type != SYM_FUNC && param != NULL){
		switch(tk->Type)
		{
		case SYM_ASSIGNSTRING:
			//K\gĕϐɓWJ
			ei->decl = TRUE;
			if(param->v->Type == TYPE_INTEGER){
				str = i2a(param->v->u.iValue);
				reg_match_value(ei, tk->nfa, str);
				mem_free(&str);
			}else{
				reg_match_value(ei, tk->nfa, param->v->u.Value);
			}
			ei->decl = FALSE;
			break;

		case SYM_DECLVARIABLE:
			//ϐݒ
			vi = DeclVariable(ei, tk->buf, tk->Err);
			if(vi == NULL){
				return NULL;
			}

			//ϐɒlݒ
			if(*tk->buf == ARGUMENT_ADDRESS){
				mem_free(&vi->v);
				vi->v = param->v;
				vi->vi = param->v->vi->vi;
			}else{
				SetValue(ei, vi->v, param->v);
				vi->vi = CopyValueList(param->vi);
			}
			break;

		default:
			Error(ei, ERR_SENTENCE, tk->Err, FALSE);
			return NULL;
		}
		param = param->next;
		while(tk->Type != SYM_FUNC && tk->Type != SYM_WORDEND){
			tk = tk->next;
		}
		if(tk->Type == SYM_WORDEND){
			tk = tk->next;
		}
	}

	if(tk != NULL && tk->Type != SYM_FUNC){
		//̈s
		ei->decl = TRUE;
		ei->to_tk = SYM_FUNC;
		if(ExecSentense(ei, tk, NULL, NULL, NULL) == -1){
			return NULL;
		}
		ei->decl = FALSE;
		ei->to_tk = 0;

		//̃XLbv
		for(; tk != NULL && tk->Type != SYM_FUNC; tk = tk->next){
			switch(tk->Type)
			{
			case SYM_EQ:
				eq = TRUE;
				break;

			case SYM_WORDEND:
				if(eq == FALSE){
					//ftHgl
					Error(ei, ERR_ARGUMENT, tk->Err, FALSE);
					return NULL;
				}
				eq = FALSE;
				break;
			}
		}
		if(eq == FALSE){
			//ftHgl
			Error(ei, ERR_ARGUMENT, tk->Err, FALSE);
			return NULL;
		}
	}
	return tk;
}


/******************************************************************************

	CopyAllArgument

	SĂ̈Azz arg ɃRs[

******************************************************************************/

static void CopyAllArgument(EXECINFO *ei, VALUEINFO *param)
{
	VALUEINFO *vi;
	VALUEINFO *To, Top;
	VALUE v;

	if(param == NULL){
		v.u.iValue = 0;
		v.Type = TYPE_INTEGER;
		SetLocalVariable(ei, ARGUMENT_VARIABLE, &v);
		return;
	}

	//QƃXg̍쐬
	To = &Top;
	To->next = NULL;
	for(; param != NULL; param = param->next){
		To = To->next = AllocValue();
		if(To == NULL){
			break;
		}
		mem_free(&To->v);
		To->v = param->v;
		To->vi = param->vi;
	}

	//Xg̒ǉ
	v.u.iValue = 0;
	v.Type = TYPE_INTEGER;
	if(SetLocalVariable(ei, ARGUMENT_VARIABLE, &v) == FALSE){
		return;
	}
	vi = GetVariable(ei, ARGUMENT_VARIABLE);
	vi->vi = Top.next;
}


/******************************************************************************

	ExecNameFunction

	Öv֐s

******************************************************************************/

static VALUEINFO *ExecNameFunction(SCRIPTINFO *csci, FUNCINFO *fi, VALUEINFO *param, long lParam)
{
	EXECINFO cei;
	VALUEINFO *vret = NULL;
	TOKEN *tk;
	int i = 0;

	ZeroMemory(&cei, sizeof(EXECINFO));
	cei.Name = fi->Name;
	cei.sci = csci;
	cei.parent = NULL;
	cei.lParam = lParam;

	//̓WJ
	tk = ExpandArgument(&cei, fi->tk->next, param);
	if(tk == NULL || tk->target == NULL){
		FreeExecInfo(&cei);
		return AllocValue();
	}
	//SĂ̈ arg ɃRs[
	CopyAllArgument(&cei, param);

	//߂l̊m
	vret = AllocValue();
	if(vret == NULL){
		Error(&cei, ERR_ALLOC, fi->Name, FALSE);
		FreeExecInfo(&cei);
		return (VALUEINFO *)-1;
	}
	//s
	if(ExecSentense(&cei, tk->target, vret, NULL, &i) == RET_ERROR){
		FreeValue(vret);
		vret = (VALUEINFO *)-1;
	}
	FreeExecInfo(&cei);
	return vret;
}


/******************************************************************************

	ExecArgumentFunction

	̃}b`֐s

******************************************************************************/

static int ExecArgumentFunction(SCRIPTINFO *sci, VALUEINFO *param, int *exit, long lParam)
{
	FUNCINFO *fi = sci->fi;
	EXECINFO ei;
	TOKEN *tk;
	int ret = 0;
	int rc;
	int i = 0;

	ZeroMemory(&ei, sizeof(EXECINFO));
	ei.sci = sci;
	ei.parent = NULL;
	ei.lParam = lParam;
	for(; fi != NULL; fi = fi->next){
		if(fi->tk == NULL || fi->tk->Type != SYM_ARGMATCH || fi->tk->next == NULL || fi->tk->next->nfa == NULL){
			continue;
		}
		tk = fi->tk->next;

		ei.decl = TRUE;
		//bZ[W̓WJ
		if(reg_match_value(&ei, tk->nfa, param->v->u.Value) == TRUE){
			ret++;
			ei.Name = fi->Name;
			ei.decl = FALSE;

			//PԖڂ̈XLbv
			for(; tk->Type != SYM_FUNC && tk->Type != SYM_WORDEND; tk = tk->next);
			if(tk->Type == SYM_WORDEND){
				tk = tk->next;
			}

			//̓WJ
			tk = ExpandArgument(&ei, tk, param->next);
			if(tk == NULL || tk->target == NULL){
				FreeExecInfo(&ei);
				break;
			}
			//SĂ̈ arg ɃRs[
			CopyAllArgument(&ei, param);

			//s
			i = 0;
			rc = ExecSentense(&ei, tk->target, NULL, NULL, &i);
			if(rc == RET_ERROR){
				FreeExecInfo(&ei);
				ret = -1;
				break;
			}
			if(rc == RET_EXIT){
				//I
				FreeExecInfo(&ei);
				if(exit != NULL){
					*exit = 1;
				}
				break;
			}
		}
		FreeExecInfo(&ei);
	}
	if(CheckEndEvent() == TRUE){
		ret = -1;
	}
	return ret;
}


/******************************************************************************

	ExecArgumentAllFunction

	SẴXNvgň̃}b`֐s

******************************************************************************/

static int ExecArgumentAllFunction(SCRIPTINFO *sci, VALUEINFO *param, long lParam)
{
	int ret = 0;
	int i = 0;
	int exit = 0;

	for(sci = sci->next; sci != NULL; sci = sci->next){
		i = ExecArgumentFunction(sci, param, &exit, lParam);
		if(i == -1){
			return -1;
		}
		ret += i;
		if(exit == 1){
			break;
		}
	}
	return ret;
}


/******************************************************************************

	ExecLibFunction

	Cu֐s

******************************************************************************/

static VALUEINFO *ExecLibFunction(EXECINFO *ei, LIBFUNC StdFunc, char *name, VALUEINFO *param)
{
	VALUEINFO *vret = NULL;
	char ErrStr[BUF_SIZE];
	int ret;

	vret = AllocValue();
	if(vret == NULL){
		Error(ei, ERR_ALLOC, name, FALSE);
		return (VALUEINFO *)-1;
	}
	*ErrStr = '\0';

	ret = StdFunc(ei, param, vret, ErrStr);
	if(ret < 0){
		switch(ret)
		{
		case -1:
			Error(ei, ERR_FUNC, ErrStr, FALSE);
			break;

		case -2:
			Error(ei, ERR_ARGUMENTCNT, name, FALSE);
			break;
		}
		FreeValue(vret);
		return (VALUEINFO *)-1;
	}
	return vret;
}


/******************************************************************************

	GetExecFuncCnt

	֐sJEg̎擾

******************************************************************************/

int GetExecFuncCnt()
{
	int ret;
	ScriptWait();
	ret = FuncCnt;
	ScriptUnWait();
	return ret;
}


/******************************************************************************

	SetExecFuncCnt

	֐sJEg̐ݒ

******************************************************************************/

static void SetExecFuncCnt(int i)
{
	ScriptWait();
	FuncCnt += i;
	ScriptUnWait();
}


/******************************************************************************

	ExecFunction

	֐s

******************************************************************************/

VALUEINFO *ExecFunction(EXECINFO *ei, TOKEN *tk, char *name, VALUEINFO *param, BOOL no_err)
{
	SCRIPTINFO *csci = ei->sci;
	VALUEINFO *vret = NULL;
	FUNCINFO *fi = ei->sci->fi;
	LIBRARYINFO *lib;
	LIBFUNC StdFunc;
	char *r;
	int name_hash;
	int i = 0;

	//ΏۃXNvǧ
	for(r = name; *r != '\0' && *r != SCRIPT_SEP; r++);
	if(*r != SCRIPT_SEP){
		r = name;
	}else if(str_cmp_n(name, ALL_ARGUMENT_EXEC, 4) == 0){
		r++;
	}else{
		r++;
		csci = GetScriptInfo(ei->sci, name);
		if(csci == NULL){
			Error(ei, ERR_SCRIPT, name, FALSE);
			return (VALUEINFO *)-1;
		}
		fi = csci->fi;
	}

	//AhXς݂̊֐s
	if(tk != NULL && tk->FuncType != 0){
		SetExecFuncCnt(1);
		switch(tk->FuncType)
		{
		case FUNC_SCRIPT:
			vret = ExecNameFunction(csci, (FUNCINFO *)tk->FuncAddr, param, ei->lParam);
			break;

		case FUNC_LIBRARY:
			vret = ExecLibFunction(ei, (LIBFUNC)tk->FuncAddr, name, param);
			break;
		}
		SetExecFuncCnt(-1);
		return vret;
	}

	//̃}b`֐s
	if(tstrcmp(r, ARGUMENT_EXEC) == 0){
		if(param == NULL){
			Error(ei, ERR_ARGUMENTCNT, ei->Err, FALSE);
			return (VALUEINFO *)-1;
		}
		SetExecFuncCnt(1);
		if(str_cmp_n(name, ALL_ARGUMENT_EXEC, 4) == 0){
			//SĂ̊֐Ŏs
			i = ExecArgumentAllFunction(csci->sci_top, param, ei->lParam);
		}else{
			i = ExecArgumentFunction(csci, param, NULL, ei->lParam);
		}
		SetExecFuncCnt(-1);
		if(i == -1){
			return (VALUEINFO *)-1;
		}
		vret = AllocValue();
		if(vret != NULL){
			//}b`񐔂߂lɂ
			vret->v->u.iValue = i;
			vret->v->Type = TYPE_INTEGER;
		}
		return vret;
	}

	//[U֐
	name_hash = str2hash(r);
	for(; fi != NULL; fi = fi->next){
		if(name_hash == fi->name_hash && tstrcmp(r, fi->Name) == 0){
			if(tk != NULL){
				//AhX̑ޔ
				tk->FuncAddr = (long)fi;
				tk->FuncType = FUNC_SCRIPT;
			}
			SetExecFuncCnt(1);
			vret = ExecNameFunction(csci, fi, param, ei->lParam);
			SetExecFuncCnt(-1);
			return vret;
		}
	}

	//Cu֐
	r = alloc_join(STD_FUNC_HEAD, name);
#ifdef NO_EXPORT
	StdFunc = GetFuncAddress(r);
#ifdef _WIN32_WCE
	{
		TCHAR *tr = char2tchar(r);
		for(lib = csci->sci_top->lib; lib != NULL && StdFunc == NULL; lib = lib->next){
			//Cǔ
			StdFunc = GetProcAddress(lib->hModul, tr);
		}
		mem_free(&tr);
	}
#else
	for(lib = csci->sci_top->lib; lib != NULL && StdFunc == NULL; lib = lib->next){
		//Cǔ
		StdFunc = GetProcAddress(lib->hModul, r);
	}
#endif
#else
	StdFunc = GetProcAddress(GetModuleHandle(NULL), r);
	for(lib = csci->sci_top->lib; lib != NULL && StdFunc == NULL; lib = lib->next){
		//Cǔ
		StdFunc = GetProcAddress(lib->hModul, r);
	}
#endif
	mem_free(&r);
	if(StdFunc == NULL){
		if(no_err == FALSE){
			Error(ei, ERR_FUNCTION, name, FALSE);
		}
		return (VALUEINFO *)-1;
	}
	if(tk != NULL){
		//AhX̑ޔ
		tk->FuncAddr = (long)StdFunc;
		tk->FuncType = FUNC_LIBRARY;
	}
	SetExecFuncCnt(1);
	vret = ExecLibFunction(ei, StdFunc, name, param);
	SetExecFuncCnt(-1);
	return vret;
}


/******************************************************************************

	ScriptThread

	XNvgsXbh

******************************************************************************/

static unsigned int CALLBACK ScriptThread(SCRIPT_THREAD_INFO *th_sci)
{
	EXECINFO ei;
	VALUEINFO *vret;

	ZeroMemory(&ei, sizeof(EXECINFO));
	ei.sci = th_sci->sci;
	ei.lParam = th_sci->lParam;
	vret = ExecFunction(&ei, th_sci->tk, th_sci->func_name, th_sci->param, th_sci->no_err);
	FreeValue(vret);

	mem_free(&(th_sci->func_name));
	FreeValueList(th_sci->param);
	mem_free(&th_sci);
	ExitThread(0);
	return 0;
}


/******************************************************************************

	CreateScriptThread

	XNvgsXbh̍쐬

******************************************************************************/

long CreateScriptThread(SCRIPTINFO *sci, TOKEN *tk, char *func_name, VALUEINFO *param, BOOL no_err, long lParam)
{
	SCRIPT_THREAD_INFO *ssci;
	long ret;

	//XNvg̃Xbh̐ݒ
	ssci = mem_alloc(sizeof(SCRIPT_THREAD_INFO));
	if(ssci == NULL){
		return FALSE;
	}
	ssci->sci = sci;
	ssci->fi = sci->fi;
	ssci->func_name = alloc_copy(func_name);
	ssci->param = CopyValueList(param);
	ssci->tk = tk;
	ssci->no_err = no_err;
	ssci->thId = -1;
	ssci->lParam = lParam;

	//Xbh̍쐬
	ssci->hth = (HANDLE)CreateThread(NULL, 0, ScriptThread,
		(void *)ssci, CREATE_SUSPENDED, (unsigned *)&ssci->thId);
	if(ssci->hth == 0){
		mem_free(&(ssci->func_name));
		FreeValueList(ssci->param);
		mem_free(&ssci);
		return -1;
	}
	//Xbh̊Jn
	ret = ssci->thId;
	ResumeThread(ssci->hth);
	return ret;
}


/******************************************************************************

	SetGlobalFlag

	O[oϐ̃tOZbg

******************************************************************************/

static void SetGlobalFlag(VALUEINFO *vi)
{
	for(; vi != NULL; vi = vi->next){
		vi->v->Global = TRUE;
		SetGlobalFlag(vi->vi);
	}
}


/******************************************************************************

	ExecGlobalArea

	O[ö̎s

******************************************************************************/

void ExecGlobalArea(SCRIPTINFO *sci)
{
	EXECINFO ei;
	int i = 0;

	SetExecFuncCnt(1);
	for(; sci != NULL; sci = sci->next){
		ZeroMemory(&ei, sizeof(EXECINFO));
		ei.Name = sci->name;
		ei.sci = sci;

		if(ExecSentense(&ei, sci->tk, NULL, NULL, &i) == -1){
			SetExecFuncCnt(-1);
			return;
		}
		SetGlobalFlag(ei.vi);
		sci->vi = ei.vi;
		ei.vi = NULL;
		FreeExecInfo(&ei);
	}
	SetExecFuncCnt(-1);
}
/* End of source */
